package comunicacao.sender;

import java.io.IOException;
import java.net.DatagramSocket;
import java.util.LinkedList;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;

import comunicacao.Socket;
import comunicacao.data.segment.Segment;
import comunicacao.data.segment.SequenceNumber;
import comunicacao.util.ConnectionState;

public class RetransmitQueue {

	private Socket socket;

	private LinkedList<Segment> retransmissionQueue;
	private Object queueMonitor;

	private ScheduledExecutorService executor;
	private ScheduledFuture<?> timer;
	private TimeoutAlgorith timeout;
	private Object timerMonitor;
	
	private int lastRetransmited;

	public RetransmitQueue(Socket socket) {
		super();
		this.socket = socket;

		this.retransmissionQueue = new LinkedList<Segment>();
		this.queueMonitor = new Object();

		this.executor = Executors.newSingleThreadScheduledExecutor();
		this.timer = null;
		this.timeout = new TimeoutAlgorith();
		this.timerMonitor = new Object();
		this.lastRetransmited =Integer.MIN_VALUE;
	}

	public void issue(Segment segment) throws IOException {
		synchronized (queueMonitor) {
			this.retransmissionQueue.add(segment);
		}
		segment.setTransmissionTime(System.currentTimeMillis());
//		System.out.println("Sending segment " + segment.getSequenceNumber().getInteger());
		this.socket.getSocket().send(segment.getPacket());

		synchronized (timerMonitor) {
			if (this.timer == null || this.timer.isDone()) {
				this.timer = executor.schedule(new Retransmit(retransmissionQueue, this.socket, this),
						this.timeout.getTimeout(), TimeUnit.MILLISECONDS);
			}
		}

	}

	public void cancelTimer() {
		synchronized (timerMonitor) {
			if (this.timer != null) {
				this.timer.cancel(false);
			}
		}

	}

	public Object getTimerMonitor() {
		return timerMonitor;
	}

	@SuppressWarnings("unused")
	private void setTimerMonitor(Object timerMonitor) {
		this.timerMonitor = timerMonitor;
	}

	public LinkedList<Segment> getRetransmissionQueue() {
		return retransmissionQueue;
	}

	public void setRetransmissionQueue(LinkedList<Segment> retransmissionQueue) {
		this.retransmissionQueue = retransmissionQueue;
	}

	public Object getQueueMonitor() {
		return queueMonitor;
	}

	@SuppressWarnings("unused")
	private void setQueueMonitor(Object monitor) {
		this.queueMonitor = monitor;
	}

	public Socket getSocket() {
		return socket;
	}

	public void setSocket(Socket socket) {
		this.socket = socket;
	}

	public ScheduledExecutorService getExecutor() {
		return executor;
	}

	public void setExecutor(ScheduledExecutorService executor) {
		this.executor = executor;
	}

	public ScheduledFuture<?> getTimer() {
		return timer;
	}

	public void setTimer(ScheduledFuture<?> timer) {
		this.timer = timer;
	}

	public TimeoutAlgorith getTimeout() {
		return timeout;
	}

	public void setTimeout(TimeoutAlgorith timeout) {
		this.timeout = timeout;
	}
	
	public int getLastRetransmited() {
		return lastRetransmited;
	}

	public void setLastRetransmited(int lastRetransmited) {
		this.lastRetransmited = lastRetransmited;
	}

	private class Retransmit implements Runnable {
		private LinkedList<Segment> retransmissionQueue;
		private Socket socket;
		private RetransmitQueue parent;

		public Retransmit(LinkedList<Segment> retransmissionQueue, Socket socket, RetransmitQueue parent) {
			super();
			this.retransmissionQueue = retransmissionQueue;
			this.socket = socket;
			this.parent = parent;
		}

		@Override
		public void run() {
			synchronized (queueMonitor) {
				Segment seg = this.retransmissionQueue.peek();
				this.parent.setLastRetransmited(seg.getSequenceNumber().getInteger() + seg.getLength());
				if (!socket.getState().equals(ConnectionState.CLOSED)) {
					synchronized (parent.getTimerMonitor()) {
						if (seg != null) {
							try {
								System.out.println(seg.getSequenceNumber().getInteger());
								this.socket.getSocket().send(seg.getPacket());
								this.parent.getTimeout().backOff();
								parent.setTimer(parent.getExecutor().schedule(
										new Retransmit(retransmissionQueue, socket, parent),
										parent.getTimeout().getTimeout(), TimeUnit.MILLISECONDS));
							} catch (IOException e) {
								// Auto-generated catch block
								// e.printStackTrace();
							}
						}
					}
				}
			}

		}

	}

	public void acknowledge() {
		synchronized (this.socket.getSeqNumMonitor()) {
			int ack = this.socket.getSendUnackSeqNum().getInteger();
			
			synchronized (queueMonitor) {
				
				if (!retransmissionQueue.isEmpty()) {
					Segment aux = retransmissionQueue.peek();
					int pos = 0;
					if (aux != null) {
						pos = aux.getSequenceNumber().getInteger() + aux.getLength();
					}
					boolean measured = false;
					while (aux != null && ((pos <= ack)
							|| (ack <= pos && this.socket.getSendNextSeqNum().getInteger() < pos))) {
						this.retransmissionQueue.remove();
						if (!measured && ack - 1 > this.lastRetransmited){
							this.timeout.rttMeasured(System.currentTimeMillis() - aux.getTransmissionTime());
							measured =true;
						}
						aux = retransmissionQueue.peek();
						if (aux != null) {
							pos = aux.getSequenceNumber().getInteger() + aux.getLength();
						} else if (this.timer != null) {
							this.cancelTimer();
						}
					}
				} else {
					if (this.timer != null) {
						this.cancelTimer();
					}
				}
			}
		}

	}
}
